/* A lot of this file is Symmetry and 80386 dependent */

#include <stdio.h>
#include <a.out.h>
#include <stab.h>
#include "util.h"
#include "quartzcommon.h"

unsigned char *core;
struct nlist *symTable;
struct nlist **symPtrs;
char *stringTable;
int numSymbols;
int numTextSymbols;

unsigned long mcountPC;
unsigned long mNumProcIdLoc;
unsigned long mEndOfTextLoc;

int mCountError = FALSE;

#define INITINFO   5000
typedef struct call_info {
  unsigned long oldLabel;
  unsigned long newLabel;
  } CallInfo;

typedef struct delay_info {
  unsigned long addr;
  unsigned long oldLabel;
  } DelayInfo;

CallInfo *callList;
DelayInfo *delayList;
int numCalls, maxCalls, numDelay, maxDelay;

#define CALL			0xe8
#define MoveIntoEax		0xb8
#define LealInto		0x8d
#define IntoEax			0x5

#define SYMB_NAME(s)	(&stringTable[(s).n_un.n_strx])
#define TEXT_END(x)		((x)->a_text - EXECPGRND)
#define DATA_END(x)		((x)->a_data + TEXT_END(x))
#define SYMB_OFF(x)		((x)->a_shdrsize + (x)->a_drsize + (x)->a_trsize + \
							(x)->a_shdata + DATA_END(x))
#define STRING_OFF(x)	((x)->a_syms + SYMB_OFF(x))

#define CHANGE(loc,val)	 (core[(loc) - EXECPGRND] = val)

int CompareByPC (p1, p2)
	struct nlist **p1, **p2;
	{
	return((*p1)->n_value - (*p2)->n_value);
	}

int SortSymbols ()
	{
	int i;

	symPtrs = MyMalloc(numSymbols, struct nlist *);
	numTextSymbols =  0;
	mEndOfTextLoc = mNumProcIdLoc = mcountPC = 0;
	for (i = 0; i < numSymbols; i++)
		{
		switch (symTable[i].n_type & N_TYPE)
			{
			case N_TEXT:
				if (!(symTable[i].n_type & N_STAB))
					symPtrs[numTextSymbols++] = &symTable[i];
				if (!strcmp(SYMB_NAME(symTable[i]), "mcount"))
					mcountPC = symTable[i].n_value - EXECPGRND;
				break;
			case N_DATA:
			case N_SHDATA:
				if (!strcmp(SYMB_NAME(symTable[i]), "_mNumProcIds"))
					mNumProcIdLoc = symTable[i].n_value - EXECPGRND;
				if (!strcmp(SYMB_NAME(symTable[i]), "_mEndOfText"))
					mEndOfTextLoc = symTable[i].n_value - EXECPGRND;
				break;
			}
		}
	if (mcountPC == 0)
		MyError("Input file needs mcount symbol", "");
	if (mNumProcIdLoc == 0 || mEndOfTextLoc == 0)
		MyError("Input file needs _mNumProcIds _mEndOfText symbols","");
	qsort((char *)symPtrs, numTextSymbols, sizeof(struct nlist *), CompareByPC);
	}

void McountError (callLoc, str)
	unsigned long callLoc;
	char *str;
	{
	int i;
	int realLoc = callLoc + EXECPGRND;

	fprintf(stderr, "Error: call mcount found at 0x%x", realLoc);
	for (i = 0; i < numTextSymbols; i++)
		if (symPtrs[i]->n_value > realLoc)
			break;
	if (i == 0)
		fprintf(stderr, " (before all text symbols):");
	else
		fprintf(stderr, " (%s):", SYMB_NAME(*symPtrs[i - 1]));
	fprintf(stderr, " %s\n", str);

	mCountError = TRUE;
	}

unsigned long GetVal (offset)
	unsigned long offset;
	{
	int i;
	unsigned long val;
	unsigned char *p = (unsigned char *)&val;

	for (i = 0; i < 4; i++)
		*(p + i) = core[offset + i];
	return(val);
	}

void PutVal (offset, val)
	unsigned long offset;
	unsigned long val;
	{
	int i;
	unsigned char *p = (unsigned char *)&val;

	for (i = 0; i < 4; i++)
		core[offset + i] = *(p + i);
	}

void FindCalls (txtEnd, mf)
	unsigned long txtEnd;
	FILE *mf;
	{
	unsigned long i, oldLabel;
	int j, oldj, k;
	char *pName;
	unsigned long numProcIds;

	fprintf(mf, "char *procNames[] = {\n");

	fprintf(mf, "\"%s\",\n\"%s\",\n\"%s\",\n\"%s\",\n",
		"Overhead", "Start", "IdleLoop", "Fork");

	numProcIds = InitProcs;
	numCalls = numDelay = 0;
	maxCalls = maxDelay = INITINFO;
	callList = MyMalloc(maxCalls, CallInfo);
	delayList = MyMalloc(maxDelay, DelayInfo);

	for (i = sizeof(struct exec), j = 0; i < txtEnd; i++)
		{
		if (core[i] == CALL && GetVal(i + 1) == (mcountPC - (i + 5)))
			{
			if (core[i - 5] != MoveIntoEax && (core[i - 6] != LealInto
						|| core[i - 5] != IntoEax))
				McountError(i, 
					"no movl $... (or leal ...) ,%eax prior to call to mcount");

			oldLabel = GetVal(i - 4);
			for (oldj = j; j < numTextSymbols; j++)
				if (i < (symPtrs[j]->n_value - EXECPGRND))
					break;
			if (oldj == j)
				{
				for (k = 0; k < numCalls; k++)
				  	if (callList[k].oldLabel == oldLabel)
						{
						PutVal(i - 4, callList[k].newLabel);
						break;
						}
				if (k == numCalls)
					{
					if (numDelay >= maxDelay)
						{
						maxDelay += INITINFO;
				    	delayList = MyReAlloc(delayList, maxDelay, DelayInfo);
						}
					delayList[numDelay].addr = i - 4;
					delayList[numDelay].oldLabel = oldLabel;
					numDelay++;
					}
				}
			else
				{
				for (k = 0; k < numDelay; k++)
					{
					if (delayList[k].addr != 0 && 
							delayList[k].oldLabel == oldLabel)
						{
						PutVal(delayList[k].addr, numProcIds);
						delayList[k].addr = 0;
						}
					}
				if (numCalls >= maxCalls)
					{
					maxCalls += INITINFO;
			    	callList = MyReAlloc(callList, maxCalls, CallInfo);
					}
				callList[numCalls].oldLabel = oldLabel;
				callList[numCalls].newLabel = numProcIds;
				numCalls++;
				pName = SYMB_NAME(*symPtrs[j - 1]);
				fprintf(mf, "\"%s\",\n",(*pName == '_') ? (pName + 1) : pName); 
				PutVal(i - 4, numProcIds++);
				}
			}
		}
	for (k = 0; k < numDelay; k++)
		{
		if (delayList[k].addr != 0)
			McountError(delayList[k].addr,
				"duplicate calls to mcount in procedure");
		}

	fprintf(mf, "};\n");
	fprintf(mf, "int numProcNames = sizeof(procNames)/sizeof(char *);\n");

	if (numProcIds <= InitProcs)
		MyError("No calls to mcount found in object file", "");
	PutVal(mNumProcIdLoc, numProcIds);
	PutVal(mEndOfTextLoc, txtEnd + EXECPGRND);
	}

char sysBuf[512];
void MakeExecutable (outfile)
	char *outfile;
	{
	sprintf(sysBuf, "chmod +x %s", outfile); 
	if (system(sysBuf))
		fprintf(stderr, "Unable to set %s to executable\n", outfile);
	}

extern char *optarg;
extern int optind;

/* The real stuff */
main (argc, argv)
	char **argv;
	{
	int c;
	char *inFile = "a.out";
	char *outFile = NULL;
	char *mFile = "procNames.c";
	FILE *inf;
	FILE *outf;
	FILE *mf;
	int inFileSize;
	struct exec *header;

	while ((c = getopt(argc, argv, "o:i:m:")) != EOF)
		switch(c)
			{
			case 'o':
				outFile = optarg;
				break;
			case 'i':
				inFile = optarg;
				break;
			case 'm':
				mFile = optarg;
				break;
			}

	if (!outFile)
		{
		outFile = MyMalloc(strlen(inFile) + 5, char);
		sprintf(outFile, "%s.p", inFile);
		}
	inf = MyOpen(inFile, "r");

	if ((inFileSize = MyFileLen(inf)) < sizeof(struct exec))
		MyError("No header in object file", inFile);
	core = MyMalloc(inFileSize, unsigned char);
	header = (struct exec *)core;
	MyRead(core, inFileSize, inf);
	MyClose(inf);

	if (header->a_magic != XMAGIC && header->a_magic != ZMAGIC)
		MyError("Illegal header in object file", inFile);

	symTable = (struct nlist *)(core + SYMB_OFF(header));
	if ((numSymbols = header->a_syms / sizeof(struct nlist)) == 0)
		MyError("No symbol table in object file", inFile);

	if (STRING_OFF(header) >= inFileSize)
		MyError("No string table in object file", inFile);
	stringTable = (char *)(core + STRING_OFF(header));

	SortSymbols();

	mf = MyOpen(mFile, "w");
	FindCalls(TEXT_END(header), mf);
	MyClose(mf);

	outf = MyOpen(outFile, "w+");
	MyWrite(core, inFileSize, outf);
	MyClose(outf);
	if (!mCountError)
		MakeExecutable(outFile);
	}
