#include <stdio.h>
#include "quartzcommon.h"
#include "quartz.h"

void free();

void CumulateBusy (b1, b2)
	BusyData *b1;
	BusyData *b2;
	{
	int i;

	for (i = 0; i < MaxEffectiveParallelism; i++)
		b1->byEffP[i] += b2->byEffP[i];
	}

static MeasuredData *CumulateData (d1, d2)
	MeasuredData *d1;
	MeasuredData *d2;
	{
	int i, j, k;

	if (!d2)
		return(d1);
	if (!d1)
		return(d2);

	CumulateBusy(&d1->busy, &d2->busy);
	for (i = 0; i < NumWho; i++)
		for (j = 0; j < NumStates; j++)
			for (k = 0; k < 2; k++)
				d1->nom.byNomP[i][j][k] += d2->nom.byNomP[i][j][k];
	for (i = 0; i < NumNumbers; i++)
		for (j = 0; j < MaxNominalParallelism; j++)
			d1->queue.length[i][j] += d2->queue.length[i][j];
	return(d1);
	}


void AggSynchList ()
	{
	int i, j, oldId;

	if (qdata.numSynch)
		qsort((char *)&dataList[qdata.numProcIds], qdata.numSynch,
				sizeof(SampleData), ObjNameCmp);

	for (i = 0, j = qdata.numProcIds - 1; i < qdata.numSynch; i++)
		{
		oldId = dataList[qdata.numProcIds + i].newId;
		if (i == 0 || ObjNameCmp(&dataList[qdata.numProcIds + i - 1],
								&dataList[qdata.numProcIds + i]))
			{
			++j;
			dataList[j] = dataList[qdata.numProcIds + i];
			dataList[j].numAgg = 1;
			dataList[j].newId = j;
			}
		else
			{
			dataList[j].numAgg++;
			dataList[j].data = CumulateData(dataList[j].data, 
										dataList[qdata.numProcIds + i].data);
			}
		qdata.synchData[oldId].myID = j;
		}

	UpdateIds();
	numTotal = j + 1;
	}

void AggKidData ()
	{
	int i, j;
	SampleData *p;

	if (qdata.numKids)
		qsort((char *)qdata.kidData, qdata.numKids, sizeof(KidData), KidCmp);

	for (i = 0, j = -1; i < qdata.numKids; i++)
		{
		if (i == 0 || KidCmp(&qdata.kidData[i-1], &qdata.kidData[i]))
			qdata.kidData[++j] = qdata.kidData[i];
		else
			CumulateBusy(&qdata.kidData[j].busy, &qdata.kidData[i].busy);
		}
	qdata.numKids = j + 1;

	for (i = 0; i < qdata.numKids; i++)
		{
		if (i == 0 || qdata.kidData[i-1].callerID != qdata.kidData[i].callerID)
			{
			p = &dataList[qdata.kidData[i].callerID];
			p->kids = &qdata.kidData[i];
			p->numKids = 1;
			}
		else
			p->numKids++;
		}
	}

void AggTriplesByCaller ()
	{
	int i, j;

	if (qdata.numTriples)
		qsort((char *)qdata.triples, qdata.numTriples,sizeof(Triple),CallerCmp);
	for (i = 0, j = -1; i < qdata.numTriples; i++)
		{
		if (i == 0 || CallerCmp(&qdata.triples[i-1], &qdata.triples[i]))
			qdata.triples[++j] = qdata.triples[i];
		else
			qdata.triples[j].num += qdata.triples[i].num;
		}
	qdata.numTriples = j + 1;
	}

int PrintTriples ()
	{
	int i;

	for (i = 0; i < qdata.numTriples; i++)
		printf("%d %d %d\n", qdata.triples[i].calleeID, 
			qdata.triples[i].callerID, qdata.triples[i].num);
	}
int PrintByCallee ()
	{
	int i;

	for (i = 0; i < qdata.numTriples; i++)
		printf("%d %d %d\n", byCallee[i].calleeID, 
			byCallee[i].callerID, byCallee[i].num);
	}
