#include <stdio.h>
#include "quartzcommon.h"
#include "quartz.h"

FLOAT tm2weight (v, i)
	FLOAT v;
	int i;
	{
	int j = i + 1;

	if (j < qdata.numProcessors)
		return(v / j);
	return(v / qdata.numProcessors);
	}

FLOAT ComputeMax (b)
	BusyData *b;
	{
	int i;
	FLOAT m = 0;

	for (i = 0; i < MaxEffectiveParallelism; i++)
		m = max(m, tm2weight(b->byEffP[i], i));
	return(m);
	}

FLOAT ComputeSum (b)
	BusyData *b;
	{
	int i;
	FLOAT total = 0;

	for (i = 0; i < MaxEffectiveParallelism; i++)
		total += tm2weight(b->byEffP[i], i);
	return(total);
	}

void ComputeWeights (s)
	SampleData *s;
	{
	int i;

	if (s->data)
		s->weight[MePlusKids] = s->weight[JustMe] = ComputeSum(&s->data->busy);
	else
		s->weight[MePlusKids] = s->weight[JustMe] = 0;
	for (i = 0; i < s->numKids; i++)
		s->weight[MePlusKids] += ComputeSum(&s->kids[i].busy);
	}

FLOAT ComputePtime (b)
	BusyData *b;
	{
	int i;
	FLOAT total = 0;

	for (i = 0; i < MaxEffectiveParallelism; i++)
		total += b->byEffP[i];
	return(total);
	}

FLOAT ComputeAvgParallelism (b)
	BusyData *b;
	{
	int i;
	FLOAT divisor, numerator;
	FLOAT w;

	divisor = numerator = 0;
	for (i = 0; i < MaxEffectiveParallelism; i++)
		{
		w = tm2weight(b->byEffP[i], i);
		divisor += w;
		numerator += (i + 1) * w;
		}
	if (divisor < EPSILON)
		return(0);
	return(numerator / divisor);
	}

void ComputeDistrib (p, b)
	SampleData *p;
	BusyData *b;
	{
	int i;

	for (i = 0; i < MaxEffectiveParallelism; i++)
		b->byEffP[i] = 0;
	if (p->data)
		CumulateBusy(b, &p->data->busy);
	for (i = 0; i < p->numKids; i++)
		CumulateBusy(b, &p->kids[i].busy);
	}

FLOAT ComputeAvgLen (q, n)
	QueueLength *q;
	int n;
	{
	int i;
	FLOAT tot = 0;
	FLOAT avg = 0;

	for (i = 0; i < MaxNominalParallelism; i++)
		tot += q->length[n][i];
	for (i = 0; i < MaxNominalParallelism; i++)
		avg += (i * q->length[n][i] / tot);
	return(avg);
	}

void ComputeStart (tripleList, which)
	Triple *tripleList;
	int which;
	{
	int i; 
	int oldID = -1;
	int newID;

	numCalls = 0;
	for (i = 0; i < qdata.numTriples; i++)
		{
		if (which == CALLER)
			newID = tripleList[i].callerID;
		else 
			newID = tripleList[i].calleeID;
		if (oldID != newID)
			{
			oldID = newID;
			dataList[oldID].graph[which].start = &tripleList[i];
			dataList[oldID].graph[which].numTriples = 1;
			}
		else
			dataList[oldID].graph[which].numTriples++;
		dataList[oldID].graph[which].numCalls += tripleList[i].num;
		numCalls += tripleList[i].num;
		}
	}

int Descend (index, dfsNum)
	int index;
	int dfsNum;
	{
	int i;
	Triple *t = dataList[index].graph[CALLER].start;

	if (dataList[index].dfsNum != 0)
		return(dfsNum);
	dataList[index].dfsNum = dfsNum++;
	for (i = 0; i < dataList[index].graph[CALLER].numTriples; i++, t++)
		dfsNum = Descend(t->calleeID, dfsNum);
	return(dfsNum);
	}

void ComputeDfs ()
	{
	int i, dfsNum;

	for (dfsNum = 1, i = 0; i < numTotal; i++)
		if (dataList[i].dfsNum == 0)
			dfsNum = Descend(i, dfsNum);
	}

void ComputeData ()
	{
	int i, j;

	for (i = 0; i < numTotal; i++)
		{
		dataList[i].dfsNum = 0;
		for (j = 0; j < NumDir; j++)
			{
			dataList[i].grap		j].start = 0;
			dataList[i].graph[j].numTriples = 0;
			dataList[i].graph[j].numCalls = 0;
			}
		ComputeWeights(&dataList[i]);
		}
	ComputeStart(qdata.triples, CALLER);
	ComputeStart(byCallee, CALLEE);
	ComputeDfs();
	}
