/*
*/
#include <stdio.h>
#include "framebuffer.h"

#define FILLED		0
#define EMPTY		1
#define POSDIAGONAL	2
#define NEGDIAGONAL	4
#define HORIZONTAL	8
#define VERTICAL	16
#define HATCH		(HORIZONTAL|VERTICAL)
#define CROSSHATCH	(NEGDIAGONAL|POSDIAGONAL)

#define WIDTH		3

#define swap(a,b)	{ int tmp; tmp = a; a = b; b = tmp; }

int ll_X, ll_Y, ur_X, ur_Y;

void NDC_clear();

void NDC_init ()
{
  if (fb_open( &ll_X, &ur_Y, &ur_X, &ll_Y) == FB_ERROR) {
    fprintf(stderr, "Can't open frame buffer: %s\n", fb_error_message);
    exit(1);
  }
  NDC_clear();
}

void NDC_close ()
{
  fb_close();
}

void NDC_clear ()
{
  fb_clear();
}

void NDC_flush ()
{
  fb_flush();
}

int ToDevX (ndc)
  double ndc;
  {
  return ((1 - ndc) * ll_X + ndc * ur_X);
  }

int ToDevY (ndc)
  double ndc;
  {
  return ((1 - ndc) * ll_Y + ndc * ur_Y);
  }

void NDCDrawText (x, y, str)
  double x, y;
  char *str;
  {
  fb_text( ToDevX(x), ToDevY(y), str);
  }

void NDCDrawLine (x1, y1, x2, y2)
  double x1, y1, x2, y2;
  {
  Bresenham(ToDevX(x1), ToDevY(y1), ToDevX(x2), ToDevY(y2), BLACK);
  }

int int2type (type)
  int type;
  {
  switch (type)
	{
	case 0: 
	  return(FILLED);
	case 1: 
	  return(EMPTY);
	case 2: 
	  return(VERTICAL);
	case 3: 
	  return(HORIZONTAL);
	case 4: 
	  return(POSDIAGONAL);
	case 5: 
	  return(NEGDIAGONAL);
	case 6: 
	  return(HATCH);
	default:
	  return(CROSSHATCH);
	}
  }

static void DrawBox (x1, y1, x2, y2, kind)
  int x1, x2, y1, y2, kind;
  {
  int i, j;
  int thisOne;

  Bresenham(x1, y1, x1, y2, BLACK);
  Bresenham(x1, y1, x2, y1, BLACK);
  Bresenham(x2, y1, x2, y2, BLACK);
  Bresenham(x1, y2, x2, y2, BLACK);

  if (x1 > x2)
	swap(x1, x2);
  if (y1 > y2)
	swap(y1, y2);

  kind = int2type(kind);
  for (i = x1; i <= x2; i++)
	for (j = y1; j <= y2; j++)
	  {
	  thisOne = 0;
	  if (kind == FILLED)
	  	thisOne = 1;
	  if ((kind & POSDIAGONAL) && (i % WIDTH) == (j % WIDTH))
		thisOne = 1;
	  if ((kind & NEGDIAGONAL) && ((ur_X - i) % WIDTH) == (j % WIDTH))
		thisOne = 1;
	  if ((kind & HORIZONTAL) && (j % WIDTH) == 0)
		thisOne = 1;
	  if ((kind & VERTICAL) && (i % WIDTH) == 0)
		thisOne = 1;

	  if (thisOne)
    	fb_writePixel( i, j, BLACK);
	  }

  }

void NDCDrawSquare (x, y, sz, kind)
  double x, y, sz;
  int kind;
  {
  int dx = (ToDevX(sz) - ToDevX((double)0.0)) / 2;
  int dy = (ToDevY(sz) - ToDevY((double)0.0)) / 2;

  if (dx > dy)
	dx = dy;
  else
	dy = dx;
  DrawBox(ToDevX(x) - dx, ToDevY(y) - dy, ToDevX(x) + dx, ToDevY(y) + dy,kind);
  }

void NDCDrawBox (x1, y1, x2, y2, kind)
  double x1, y1, x2, y2; 
  {
  DrawBox(ToDevX(x1), ToDevY(y1), ToDevX(x2), ToDevY(y2), kind);
  }

/*
** Draw a line from (x1,y1) to (x2,y2) using Bresenham's algorithm, assuming
** that x1 < x2, y1 < y2,  and that (y2 - y1) <= (x2 - x1). All coordinates
** are assumed to be expressed in raw device coordinates.
*/
Bresenham( x1, y1, x2, y2, color)
int x1, y1, x2, y2;		/* The coordinates of the endpoints */
COLOR color;			/* The color to draw the line in */
{
  int d, dx, dy, incr1, incr2, x, y, xincr, yincr;
  int reverse;

  dx = x2 - x1;
  dy = y2 - y1;
  fb_writePixel( x1, y1, color);
  fb_writePixel( x2, y2, color);

  reverse = 0;
  xincr = 1;
  yincr = 1;
  if (dx < 0)
	{
	xincr = -1;
	dx = -dx;
	}
  if (dy < 0)
	{
	yincr = -1;
	dy = -dy;
	}
  if (dy > dx)
	{
	swap(x1,y1);
	swap(x2,y2);
	swap(dx,dy);
	swap(xincr,yincr);
	reverse = 1;
	}

  /* Do necessary computation of loop-invariant quantities */
  incr1 = dy << 1;
  d = incr1 - dx;
  incr2 = d - dx;

  /* Start at the first endpoint (x1,y1) */
  x = x1; y = y1;

  /* Paint the rest of pixels on the line */
  while (x != x2) {
    x += xincr;
    if (d < 0) {
      /* Don't move up */
      d += incr1;
    } else {
      /* Do move up */
      d += incr2;
      y += yincr;
    }
    if (reverse)
    	fb_writePixel( y, x, color);
    else
    	fb_writePixel( x, y, color);
  }
}
