#include <stdio.h>
#include "quartzcommon.h"
#include "quartz.h"

/* Lots of magic; I don't know if these will look okay. */
#define MaxBoxes	8

#define TICKLEN 	.008
#define NUMTICKS	10

#define EPSWIDTH	.008
#define TEXTHEIGHT	.012

typedef struct point {
	double x;
	double y;
	} POINT;

POINT hdrText = { .4, .96};
POINT origin = { .14, .50 };
POINT axisLen = { .85, .44 };

POINT xLabel = { .4, .44 };
POINT yLabel = { .02, .96 };
POINT tickLabel = { .02, .47};

POINT entryBox = { .4, .40};
POINT entryText = { .45, .40};

double boxSize = .015;
double deltaEntry = .04;

char tickBuf[80];
char entryBuf[80];
char readBuf[80];

/* Reset each time we display a new procedure */
double sofar[MaxEffectiveParallelism];
double yScale;
int entryCount;

static char *skipspace (p)
	char *p;
	{
	while (*p && (*p == ' ' || *p == '\t' || *p == '\n'))
		p++;
	return(p);
	}

static void SetYScale (w)
	FLOAT w;
	{
	double ceil(); 

	yScale = 1;
	while (w >= 10)
		{
		yScale *= 10;
		w /= 10;
		}
	yScale *= ceil((double)w);
	}

static void FillEntry (name, id)
	char *name;
	int id;
	{
	sprintf(entryBuf, "%s [%d]", name, id + 1);
	}

static void DisplayFrame (p, w)
	SampleData *p;
	FLOAT w;
	{
	int i;
	double units;
	POINT tick;

	NDC_clear();
	sprintf(entryBuf, "Quartz: %s [%d]", p->name, p->newId + 1);
	NDCDrawText(hdrText.x, hdrText.y, entryBuf);

	NDCDrawLine(origin.x, origin.y, origin.x + axisLen.x, origin.y);
	NDCDrawLine(origin.x, origin.y, origin.x, origin.y + axisLen.y);

	SetYScale(w);
	units = yScale / 10;
	if (yScale < 100000)
		NDCDrawText(yLabel.x, yLabel.y, "Weight (us.)");
	else
		{
		units /= 1000000;
		NDCDrawText(yLabel.x, yLabel.y, "Weight (sec.)");
		}
	NDCDrawText(xLabel.x, xLabel.y, "number of processors");

	for (i = 0; i <= NUMTICKS; i++)
		{
		tick.y = origin.y + (i * axisLen.y) / NUMTICKS ;
		sprintf(tickBuf, "%8.2#f", units * i);
		NDCDrawText(tickLabel.x, tick.y - TEXTHEIGHT / 2, tickBuf);
		if (i != 0)
			NDCDrawLine(origin.x - TICKLEN, tick.y, origin.x, tick.y);
		}
	for (i = 0; i < qdata.numProcessors; i++)
		{
		tick.x = origin.x;
		tick.x += ((i + .5) * axisLen.x) / qdata.numProcessors; 
		sprintf(tickBuf, "%d", i + 1);
		NDCDrawText(tick.x, tickLabel.y, tickBuf);
		sofar[i] = origin.y;
		}
	entryCount = 0;
	}

static void DisplayEntry (name, id, data)
	char *name;
	int id;
	BusyData *data;
	{
	POINT ll, ur, entry;
	int i;

	for (i = 0; i < qdata.numProcessors; i++)
		{
		ll.x = origin.x + (i * axisLen.x / qdata.numProcessors) + EPSWIDTH; 
		ur.x = origin.x + ((i + 1) * axisLen.x / qdata.numProcessors) - EPSWIDTH; 

		ll.y = sofar[i];
		ur.y = ll.y + tm2weight(data->byEffP[i], i) * axisLen.y / yScale ;
		sofar[i] = ur.y;
		NDCDrawBox(ll.x, ll.y, ur.x, ur.y, entryCount);
		}
	entry.y = entryText.y - (entryCount * deltaEntry);
	FillEntry(name, id);
	NDCDrawText(entryText.x, entry.y - TEXTHEIGHT/2, entryBuf);
	NDCDrawSquare(entryBox.x, entry.y, boxSize, entryCount);
	entryCount++;
	}

static void DisplayProcedure (p)
	SampleData *p;
	{
	int i, j, maxboxes;
	SampleData *q;
	CallInfo *c;
	BusyData w;
	LinkData *linkList;

	c = &p->graph[CALLER];
	if (c->numTriples)
		linkList = FillLink(c);

	ComputeDistrib(p, &w);
	DisplayFrame(p, ComputeMax(&w));

	maxboxes = MaxBoxes - 1;
	if (p->data)
		{
		DisplayEntry(p->name, p->newId, &p->data->busy);
		maxboxes--;
		}
	maxboxes = min(c->numTriples, maxboxes);
	for (i = 0; i < maxboxes; i++)
		{
       	q = &dataList[linkList[i].triple.calleeID];
		DisplayEntry(q->name, q->newId, linkList[i].data);
		}
	if (i < c->numTriples)
		{
		for (j = 0; j < MaxEffectiveParallelism; j++)
			w.byEffP[j] = 0;
		for (; i < c->numTriples; i++)
			CumulateBusy(&w, linkList[i].data);
		DisplayEntry("other", 0, &w);
		}
	NDC_flush();
	}

void Display ()
	{
	char *str;
	int id;

	NDC_init();
	while (TRUE)
		{
		printf("Quartz>");
		str = skipspace(gets(readBuf));
		if (*str == '\0')
			continue;
		if (!strcmp(str, "quit"))
			break;
		if (str[0] < '0' || str[0] > '9')
			printf("Not a number, type quit to exit.\n");
		else if ((id = atoi(str) - 1) >= numTotal)
			printf("No such id: %d\n", id);
		else if (sorted[MePlusKids][id]->weight[MePlusKids] < EPSILON)
			printf("%s has no weight to display\n", sorted[MePlusKids][id]->name);
		else
			DisplayProcedure(sorted[MePlusKids][id]);
		}
	NDC_close();
	}

