#include <stdio.h>
#include "quartzcommon.h"
#include "quartz.h"
#include "util.h"

SampleData *dataList;
SampleData **sorted[NumWho];
Triple *byCallee;
int numTotal;
int numCalls;

int id2newid (id)
	int id;
	{
	if (id >= qdata.numProcIds)
		return(qdata.synchData[id - qdata.numProcIds].myID);
	return(id);
	}

void UpdateIds ()
	{
	int i;

	for (i = 0; i < qdata.numTriples; i++)
		{
		qdata.triples[i].calleeID = id2newid(qdata.triples[i].calleeID);
		qdata.triples[i].callerID = id2newid(qdata.triples[i].callerID);
		}
	for (i = 0; i < qdata.numKids; i++)
		{
		qdata.kidData[i].callerID = id2newid(qdata.kidData[i].callerID);
		qdata.kidData[i].calleeID = id2newid(qdata.kidData[i].calleeID);
		}
	}

int ObjNameCmp (o1, o2)
	SampleData *o1;
	SampleData *o2;
	{
	int r = strcmp(o1->name, o2->name);

	if (r)
		return(r);
	if (r = strcmp(o1->parentName, o2->parentName))
		return(r);
	return(o1->kind - o2->kind);
	}

int KidCmp (k1, k2)
	KidData *k1, *k2;
	{
	int rv = k1->callerID - k2->callerID;

	if (rv)
		return(rv);
	return(k1->calleeID - k2->calleeID);
	}

int CallerCmp (g1, g2)
	Triple *g1, *g2;
	{
	int rv = g1->callerID - g2->callerID;

	if (rv)
		return(rv);
	return(g1->calleeID - g2->calleeID);
	}

int CalleeCmp (g1, g2)
	Triple *g1, *g2;
	{
	int rv = g1->calleeID - g2->calleeID;

	if (rv)
		return(rv);
	return(g1->callerID - g2->callerID);
	}

void SortByCallee ()
	{
	int i;

	byCallee = MyMalloc(qdata.numTriples, Triple);
	for (i = 0; i < qdata.numTriples; i++)
		byCallee[i] = qdata.triples[i];
	qsort((char *)&byCallee[0], qdata.numTriples, sizeof(Triple), CalleeCmp);
	}

/* more calls first */
int CallsCmp (o1, o2)
	SampleData *o1;
	SampleData *o2;
	{
	int r = o2->graph[CALLEE].numCalls - o1->graph[CALLEE].numCalls;

	if (r)
		return(r);
	return(ObjNameCmp(o1, o2));
	}

int FLOATCmp (d1, d2)
	FLOAT d1, d2;
	{
	FLOAT r = d1 - d2;

	if ((r > SIGNIFICANT) || (r < -SIGNIFICANT))
		{
		if (r < 0)
			return(-1);
		else
			return(1);
		}
	return(0);
	}

/* bigger weight first */
int SelfCmp (o1, o2)
	SampleData **o1;
	SampleData **o2;
	{
	int r = FLOATCmp((*o2)->weight[JustMe], (*o1)->weight[JustMe]);

	if (r != 0)
		return(r);
	return(CallsCmp(*o1, *o2));
	}

/* but smaller dfsNum first */
int TotalCmp (o1, o2)
	SampleData **o1;
	SampleData **o2;
	{
	int r = FLOATCmp((*o2)->weight[MePlusKids], (*o1)->weight[MePlusKids]);

	if (r != 0)
		return(r);
	r = (*o1)->dfsNum - (*o2)->dfsNum;
	if (r != 0)
		return(r);
	return(CallsCmp(*o1, *o2));
	}

int SortByWeight ()
	{
	int i, j;

	for (i = 0; i < NumWho; i++)
		sorted[i] = MyMalloc(numTotal, SampleData *);

	for (i = 0; i < numTotal; i++)
		for (j = 0; j < NumWho; j++)
			sorted[j][i] = &dataList[i];

	qsort((char *)sorted[JustMe], numTotal, sizeof(SampleData *), SelfCmp);
	qsort((char *)sorted[MePlusKids], numTotal, sizeof(SampleData *), TotalCmp);

	for (i = 0; i < numTotal; i++)
		sorted[MePlusKids][i]->newId = i;
	}

BusyData noData = {0};
static BusyData *FindData (caller, callee)
	int caller, callee;
	{
	int i;

	for (i = 0; i < dataList[caller].numKids; i++)
		if (dataList[caller].kids[i].calleeID == callee)
			return(&dataList[caller].kids[i].busy);
	return(&noData);
	}

static int LinkCmp (l1, l2)
	LinkData *l1, *l2;
	{
	int r = FLOATCmp(l2->weight, l1->weight);

	if (!r)
		r = l2->triple.num - l1->triple.num;
	if (!r)
		r = ObjNameCmp(&dataList[l1->triple.callerID], 
					&dataList[l2->triple.callerID]);
	if (!r)
		r = ObjNameCmp(&dataList[l1->triple.calleeID], 
					&dataList[l2->triple.calleeID]);
	return(r);
	}

LinkData *FillLink (c)
	CallInfo *c;
	{
	int i;
	LinkData *linkList;

	linkList = MyMalloc(c->numTriples, LinkData);
	for (i = 0; i < c->numTriples; i++)
		{
		linkList[i].triple = c->start[i];
		linkList[i].data =
			FindData(c->start[i].callerID, c->start[i].calleeID);
		linkList[i].weight = ComputeSum(linkList[i].data);
		}
	qsort((char *)linkList, c->numTriples, sizeof(LinkData), LinkCmp);
	return(linkList);
	}

void main (argc, argv)
	int argc;
	char **argv;
	{
	int i, j, id;

	numTotal = qdata.numProcIds + qdata.numSynch;
	if (qdata.numProcIds != numProcNames)
		{
		printf("Procedure name file not consistent with data file\n");
		exit(1);
		}

	dataList = MyMalloc(numTotal, SampleData);
	for (i = 0; i < qdata.numProcIds; i++)
		{
		dataList[i].name = procNames[i];
		dataList[i].kind = VarProcedure;
		dataList[i].data = NULL;
		dataList[i].numAgg = 0;
		dataList[i].parentName = "";
		}
	for (i = 0; i < qdata.numProcs; i++)
		dataList[qdata.procData[i].id].data = &qdata.procData[i].meas;

	for (i = qdata.numProcIds, j = 0; j < qdata.numSynch; i++, j++)
		{
		dataList[i].name = qdata.synchData[j].name;
		dataList[i].kind = qdata.synchData[j].kind;
		id = qdata.synchData[j].parentID;
		if (id > qdata.numProcIds)
			dataList[i].parentName = qdata.synchData[id - qdata.numProcIds].name;
		else
			dataList[i].parentName = procNames[id];
		if (qdata.synchData[j].hit)
			dataList[i].data = &qdata.synchData[j].meas;
		else
			dataList[i].data = NULL;
		dataList[i].newId = j;
		}

	AggSynchList();
	AggKidData();
	AggTriplesByCaller();
	SortByCallee();

	ComputeData();
	qdata.timeDiff -= dataList[0].weight[JustMe];

	SortByWeight();

	if (argc > 1 && ((ofd = MyOpen(argv[1], "w")) == NULL))
		{
		fprintf(stderr, "Unable to open %s\n", argv[1]);
		return;
		}
	GenReport();
	if (argc > 1)
		Display();
	}
