#include <stdio.h>
#include "quartzcommon.h"
#include "quartz.h"

#define WEIGHTLEN	4
#define CUMSECLEN	7
#define SECONDSLEN	7
#define CALLSLEN	5

FILE *ofd = stdout;

static FLOAT cumsec;
static int timeSize, callSize;

static char nameBuf[80];
static char hdrBuf[80];
static char flatBuf[80];

static char spontBuf[80];
static char linkBuf[80];
static char selfBuf1[80];
static char selfBuf2[80];

/* Some utility routines */

static FLOAT usec2sec (u)
	FLOAT u;
	{
	return(u / 1000000);
	}

static FLOAT usec2perc (u)
	FLOAT u;
	{
	return (u / (qdata.timeDiff / 100));
	}

static int int2Size (n)
	int n;
	{
	int count;

	for (count = 1; n; count++)
		n /= 10;
	return(count);
	}

static int float2size (n, sigbits)
	FLOAT n;
	int sigbits;
	{
	int count;

	for (count = 1; n >= 1.0; count++)
		n /= 10;
	return(count + sigbits + 1);
	}

static int PrintFunction (p)
	SampleData *p;
	{
	return(p->weight[MePlusKids] > EPSILON || p->graph[CALLEE].numTriples);
	}

static void FillName (p, verbose)
	SampleData *p;
	int verbose;
	{
	char *s;

	switch (p->kind)
		{
		case VarSpinLock:
			s = "spin lock";
			break;
		case VarSpinDependency:
			s = "spin dep'cy";
			break;
		case VarBlockLock:
			s = "blocking lock";
			break;
		case VarBlockDependency:
			s = "blocking dep'cy";
			break;
		case VarThread:
			s = "thread";
			break;
		}
	if (p->kind == VarProcedure || !verbose)
		sprintf(nameBuf, "%s", p->name);
	else
		sprintf(nameBuf, "%s (%s init'ed in %s)", p->name, s, p->parentName);
	}

/* Generate header */

static void GenHeader ()
	{
	FLOAT s = usec2sec(qdata.timeDiff);

	fprintf(ofd, "\nTime (seconds): %5.2#f\n", s);
	fprintf(ofd, "Samples per processor-second: %5.2#f\n", qdata.numSamples / s);
	fprintf(ofd, "Processors: %d\n\n\n", qdata.numProcessors);

/*
	fprintf(ofd, "Number of sampled procedures: %d\n", qdata.numProcIds);
	fprintf(ofd, "Number of sampled objects: %d, aggregated to %d\n", 
				qdata.numSynch, numTotal - qdata.numProcIds);
*/
	}

/* Generate flat profile part of output */

static void GenFlatItem (p)
	SampleData *p;
	{
	FLOAT w = p->weight[JustMe];
	FLOAT s = usec2sec(w);

	cumsec += s;
	FillName(p, TRUE);
	fprintf(ofd, flatBuf, usec2perc(w), cumsec, s, 
		p->graph[CALLEE].numCalls, nameBuf);
	}

static void GenFlat ()
	{
	int i;

	GenFlatDoc();

	timeSize = float2size(usec2sec(qdata.timeDiff), 2);
	callSize = int2Size(numCalls);
	callSize = max(callSize, CALLSLEN);

	sprintf(hdrBuf, "%%%ds %%%ds %%%ds %%%ds %%s\n", WEIGHTLEN, 
		max(CUMSECLEN, timeSize), max(SECONDSLEN, timeSize), callSize);

	fprintf(ofd, hdrBuf, "%NPT", "cumsecs", "seconds", "calls", "name");

	sprintf(flatBuf, "%%%d.1#f %%%d.2#f %%%d.2#f %%%dd %%s\n", WEIGHTLEN, 
		max(CUMSECLEN, timeSize), max(SECONDSLEN, timeSize), callSize);

	for (cumsec = 0, i = 0; i < numTotal; i++)
		if (PrintFunction(sorted[JustMe][i]))
			GenFlatItem(sorted[JustMe][i]);
	}

/* Generate Call Graph */

static void GenWeights (p, w)
	SampleData *p;
	BusyData *w;
	{
	int i;
	FLOAT weight, self;

	fprintf(ofd, "\nDistribution of parallelism: total (self)\n");
	for (i = 0; i < MaxEffectiveParallelism; i++)
		{
		self = 0;
		if (p->data)
			self = tm2weight(p->data->busy.byEffP[i], i);
		weight = tm2weight(w->byEffP[i], i);
		if (weight > SIGNIFICANT)
			{
			fprintf(ofd,"   %d: %.2#f %.2#f%%", i + 1,
				usec2sec(weight), usec2perc(weight));
			if (self > SIGNIFICANT)
				fprintf(ofd," (%.2#f %.2#f%%)",usec2sec(self),usec2perc(self));
			fprintf(ofd, "\n");
			}
		}
	fprintf(ofd, "\n");
	}

static void GenNomData (p)
	NominalData *p;
	{
	char *s, *t;
	int i, j, first = TRUE;
	FLOAT w[2];

	for (i = 0; i < NumWho; i++)
		{
		if (i == MePlusKids)
			s = "MePlusKids";
		else
			s = "JustMe";
		for (j = 0; j < NumStates; j++)
			{
			switch (j)
				{
				case BUSY:
					t = "busy";
					break;
				case SPIN:
					t = "spin";
					break;
				case BLOCKED:
					t = "blocked";
					break;
				case READY:
					t = "ready";
					break;
				}
			w[0] = p->byNomP[i][j][0];
			w[1] = p->byNomP[i][j][1];
			if (w[0] + w[1] > SIGNIFICANT)
				{
				if (first)
					{
					fprintf(ofd, "Nominal Data (raw processor time):\n");
					first = FALSE;
					}
			  	fprintf(ofd,"  %s %s: %.2#f < P <= %.2#f\n",
					s, t, usec2sec(w[0]), usec2sec(w[1]));
				}
			}
		}
	if (first == 0)
		fprintf(ofd, "\n");
	}

static void GenQueue (q, kind)
	QueueLength *q;
	int kind;
	{
	int i, j, pos;
	int numN;
	FLOAT avg;

	if (kind == VarSpinLock || kind == VarBlockLock)
		numN = 1;
	else
		numN = 2;
	for (i = 0; i < numN; i++)
		{
		avg = ComputeAvgLen(q, i);
		fprintf(ofd, "Average queue length");
		if (numN == 2)
			{
			if (i == 0) 
				fprintf(ofd, " (checked in)");
			else
				fprintf(ofd, " (waiting)");
			}
		fprintf(ofd, ": %.2#f\n", avg);
		if (avg > 0.01)
			for (pos = 0, j = 0; j < MaxNominalParallelism; j++) 
				{
				if (q->length[i][j] > SIGNIFICANT)
					{
					if (pos != 0)
						fprintf(ofd, ", ");
					fprintf(ofd, "%d: %.3#f", j, usec2sec(q->length[i][j]));
					if (++pos >= 7)
						{
						fprintf(ofd, "\n");
						pos = 0;
						}
					}
				}
		}
	fprintf(ofd, "\n\n");
	}

static void GenNumeric (p)
	SampleData *p;
	{
	BusyData w;

	ComputeDistrib(p, &w);

	FillName(p, TRUE);
	fprintf(ofd, "Name: %s\n", nameBuf);
	fprintf(ofd, "Index: [%d]\n\n", p->newId + 1);

	if (p->numAgg > 1)
		fprintf(ofd,"Aggregation of %d physical synch. objects\n\n", p->numAgg);

	fprintf(ofd, "NPT: %.2#f  %.2#f%%, Processor Time: %.2#f\n", 
		usec2sec(p->weight[MePlusKids]), usec2perc(p->weight[MePlusKids]), 
		usec2sec(ComputePtime(&w)));

	if (p->weight[MePlusKids] > EPSILON)
		{
		fprintf(ofd, "Average parallelism: %.2#f\n", ComputeAvgParallelism(&w));
		GenWeights(p, &w);
		}
	if (p->data != NULL)
		{
		GenNomData(&p->data->nom);
		if (p->kind != VarProcedure)
			GenQueue(&p->data->queue, p->kind);
		}
	}

static void GenCallers (callee)
	SampleData *callee;
	{
	int i;
	CallInfo *c = &callee->graph[CALLEE];
	LinkData *linkList;
	SampleData *p;

	if (c->numCalls == 0)
		{
		fprintf(ofd, spontBuf, "");
		fprintf(ofd, "\n");
		return;
		}
	linkList = FillLink(c);
	for (i = 0; i < c->numTriples; i++)
		{
		p = &dataList[linkList[i].triple.callerID];
		FillName(p, FALSE);
		fprintf(ofd, linkBuf, usec2sec(linkList[i].weight), 
			linkList[i].triple.num, c->numCalls, nameBuf, p->newId + 1);
		}
	free((char *)linkList);
	fprintf(ofd, "\n");
	}

static void GenSelf (self)
	SampleData *self;
	{
	int i;
	Triple *t = self->graph[CALLER].start;
	int selfCalls;

	for (selfCalls = 0, i = 0; i < self->graph[CALLER].numTriples; i++)
		if (t->calleeID == t->callerID)
			{
			selfCalls = t->num;
			break;
			}
	FillName(self, FALSE);
	if (selfCalls)
		fprintf(ofd, selfBuf1, usec2sec(self->weight[JustMe]), 
				usec2sec(self->weight[MePlusKids]), selfCalls, 
				self->graph[CALLEE].numCalls, nameBuf, self->newId + 1);
	else
		fprintf(ofd, selfBuf2, usec2sec(self->weight[JustMe]),
				usec2sec(self->weight[MePlusKids]),
				self->graph[CALLEE].numCalls, nameBuf, self->newId + 1);
	}

static void GenCallees (caller)
	SampleData *caller;
	{
	int i;
	CallInfo *c = &caller->graph[CALLER];
	LinkData *linkList;
	SampleData *p;

	if (c->numCalls == 0)
		return;
	linkList = FillLink(c);
	fprintf(ofd, "\n");
	for (i = 0; i < c->numTriples; i++)
		{
		p = &dataList[linkList[i].triple.calleeID];
		FillName(p, FALSE);
		fprintf(ofd, linkBuf, usec2sec(linkList[i].weight), 
			linkList[i].triple.num, p->graph[CALLEE].numCalls,
			nameBuf, p->newId + 1);
		}
	free((char *)linkList);
	}

static void GenGraph (self)
	SampleData *self;
	{
	fprintf(ofd, "Call graph:\n\n");
	sprintf(hdrBuf, "%%%ds   %%%ds/%%%ds   %%s\n", 2 * timeSize + 1,
		callSize, callSize);
	fprintf(ofd, hdrBuf, "NPT", "calls", "total", "name");

	sprintf(spontBuf, "%%%ds   <spontaneous>\n", 
		2 * timeSize + 2 * callSize + 5);
	sprintf(linkBuf, "%%%d.2#f   %%%dd/%%%dd   %%s [%%d]\n",
			2 * timeSize + 1, callSize, callSize);
	sprintf(selfBuf1, "%%%d.2#f/%%%d.2#f   %%%dd/%%%dd   %%s [%%d]\n",
			timeSize, timeSize, callSize, callSize);
	sprintf(selfBuf2, "%%%d.2#f/%%%d.2#f   %%%dd   %%s [%%d]\n",
			timeSize, timeSize, 2 * callSize + 1);


	GenCallers(self);
	GenSelf(self);
	GenCallees(self);
	fprintf(ofd, "\n-----------------------------------------------\n\n");
	}

static void GenDetailed ()
	{
	int i;
	SampleData *p;

	GenGraphDoc();
	for (i = 0; i < numTotal; i++)
		{
		p = sorted[MePlusKids][i];
		if (PrintFunction(p))
			{
			GenNumeric(p);
			GenGraph(p);
			}
		}
	}

void GenReport ()
	{
	GenHeader();
	GenFlat();
	GenDetailed();
	fflush(ofd);
	}
